"""
DropIt
Copyright: 2015.08.30 Robert Templeton
Written for CINEMA 4D R13.058 and later
"""

import os
import sys
import c4d
from c4d import plugins, utils, bitmaps, gui

# Be sure to use a unique ID obtained from www.plugincafe.com
PLUGIN_ID = 1034734

# DIDialog ID
ID_DIDIALOG = 10000

# gui elements
DID_SSUBD = 5000
DID_DSUBD = 5001
DID_OFFSET = 5002
DID_DROPIT = 5003
DID_SOURCE = 5004
DID_WHICHTARGET = 5005
DID_TARGET = 5006
DID_SOURCE_SET = 5007
DID_TARGET_SET = 5008
TARGET_FLOOR = 0
TARGET_WORKPLANE = 1
TARGET_OBJECT = 2

# Constants
BIGPOSREAL =	1000000000000.0
BIGNEGREAL =	-BIGPOSREAL

#*****************************************************************************
class VMinMax:
	# Data
	# - Reals
	minX = 0.0
	minZ = 0.0
	maxX = 0.0
	maxZ = 0.0
#*****************************************************************************

#*****************************************************************************
class Triangle:
	# Data
	# - All Vector
	# -- Triangle Vertices
	a = None
	b = None
	c = None
	# -- Triangle Centroid
	ctr = None
	# -- -Y Directed rays for above
	ra = None
	rb = None
	rc = None
	rctr = None

	# Initialize Vectors from inputs
	#*---------------------------------------------------------------------------*
	def CopyFrom(self,va,vb,vc,len):
	# - Vector va,vb,vc, Real len
	#*---------------------------------------------------------------------------*
		self.a =	c4d.Vector(va.x,va.y,va.z)
		self.ra =	c4d.Vector(va.x,va.y+len,va.z)
		self.b = 	c4d.Vector(vb.x,vb.y,vb.z)
		self.rb =	c4d.Vector(vb.x,vb.y+len,vb.z)
		self.c = 	c4d.Vector(vc.x,vc.y,vc.z)
		self.rc =	c4d.Vector(vc.x,vc.y+len,vc.z)
		self.ctr =	(self.a + self.b + self.c) * 0.33333333
		self.rctr =	c4d.Vector(self.ctr.x,self.ctr.y+len,self.ctr.z)
	#*---------------------------------------------------------------------------*
#*****************************************************************************

#*****************************************************************************
class Process:
	# Data
	# - Bool
	intersect = False
	# - Real
	minY = 0.0
	# - VMinMax list
	dbounds = None
	# - Real
	yhitpos = 0.0
	

	#*---------------------------------------------------------------------------*
	def PrintMatrix(self,m):
	# - Matrix m
	#*---------------------------------------------------------------------------*
		print "off: ",m.off.x,",",m.off.y,",",m.off.z
		print "v1 : ",m.v1.x,",",m.v1.y,",",m.v1.z
		print "v2 : ",m.v2.x,",",m.v2.y,",",m.v2.z
		print "v3 : ",m.v3.x,",",m.v3.y,",",m.v3.z
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Start_Floor(self,doc,src,offset,ssubd):
	# - BaseDocument doc, BaseList2D List src, LONG offset, LONG ssubd
	#*---------------------------------------------------------------------------*
		doc.StartUndo()
		for i, obj in enumerate(src):
			# Convert to points
			spop = self.Pointize(src[i], ssubd)
			if spop is None:
				print "Error: DropIt.Process.Start_Floor.spop ("+src[i].GetName()+")"
				continue
			self.DropToFloor(doc, src[i], spop, offset)
		doc.EndUndo()
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Start_Workplane(self,doc,src,offset,ssubd):
	# - BaseDocument doc, BaseList2D List src, LONG offset, ssubd
	#*---------------------------------------------------------------------------*
		# Must be in Workplane Mode
		#mode = doc.GetMode()
		#if mode != c4d.Mworkplane:
		#	MessageDialog("Must be in Workplane Mode!")
		#	return False
		# Get Workplane object
		wpop =	c4d.modules.snap.GetWorkplaneObject(doc)
		if wpop is None:
			print "Error: DropIt.Process.Start_Workplane.wpop"
			return False
		# Get Workplane global matrix
		wpm =	wpop.GetMg()
		iwpm =	wpm.__invert__()
		#print "Workplane Global Matrix:"
		#self.PrintMatrix(mg)
		doc.StartUndo()
		for i, obj in enumerate(src):
			# Convert to points
			spop = self.Pointize(src[i], ssubd)
			if spop is None:
				print "Error: Process.Start_Workplane.spop ("+src[i].GetName()+")"
				continue
			self.DropToWorkplane(doc, src[i], spop, wpm, iwpm, offset)
		doc.EndUndo()
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Start_Object(self,doc,src,dst,offset,ssubd,dsubd):
	# - BaseDocument doc, BaseList2D List src, dst, LONG offset, ssubd, dsubd
	#*---------------------------------------------------------------------------*
		dpop = self.Pointize_Dst(dst, dsubd)
		if not (dpop and dpop.CheckType(c4d.Opolygon)):
			print "Error: DropIt.Process.Start_Object.dpop"
			return False
		dpcnt =	dpop.GetPolygonCount()
		if not dpcnt:
			print "Error: DropIt.Process.Start_Object.dpcnt"
			return False
		self.dbounds = []
		dpoints = dpop.GetAllPoints()
		dpoly = dpop.GetAllPolygons()
		for i, obj in enumerate(dpoly):
			if dpoly[i].c == dpoly[i].d:
				vmm = self.GetPolyBoundingRect_Tri(dpoints[dpoly[i].a], dpoints[dpoly[i].b], dpoints[dpoly[i].c])
				self.dbounds.append(vmm)
			else:
				vmm = self.GetPolyBoundingRect_Quad(dpoints[dpoly[i].a], dpoints[dpoly[i].b], dpoints[dpoly[i].c], dpoints[dpoly[i].d])
				self.dbounds.append(vmm)
		doc.StartUndo()
		for i, obj in enumerate(src):
			# Convert to points
			spop = self.Pointize(src[i], ssubd)
			if spop is None:
				print "Error: DropIt.Process.Start_Object.spop ("+src[i].GetName()+")"
				continue
			self.DropToTarget(doc, src[i], spop, dpop, offset)
		doc.EndUndo()
		return True
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Pointize_Dst(self, darray, subd):
	# - BaseList2D List darray, LONG subd
	#*---------------------------------------------------------------------------*
		# For Join command
		ndoc = c4d.documents.BaseDocument()
		if ndoc is None:
			return None
		nop = c4d.BaseObject(c4d.Onull)
		if nop is None:
			return None
		ndoc.InsertObject(nop)
		pop = None
		# MCOMMAND_CURRENTSTATETOOBJECT
		for i, obj in enumerate(darray):
			pop = self.Command_CSTO(ndoc, darray[i])
			if pop is None:
				return None
			pop.InsertUnderLast(nop)
			# Get Global matrix transforms
			mg = darray[i].GetUpMg()
			vadr = pop.GetAllPoints()
			if vadr is None:
				return None
			for v, obj in enumerate(vadr):
				vadr[v] = vadr[v] * mg

		# MCOMMAND_JOIN: Use temporary doc and Null object for grouping
		pop = self.Command_Join(ndoc, nop)
		if pop is None:
			return None

		# MCOMMAND_SUBDIVIDE
		if subd != 0:
			if self.Command_Subdivide(pop, subd) is False:
				return None

		pop.Remove()
		return pop
	#*---------------------------------------------------------------------------*

	# Make Object, including hierarchy, into PointObject (for Polygons and Splines)
	#*---------------------------------------------------------------------------*
	def Pointize(self, op, subd):
	# - BaseObject op, LONG subd
	#*---------------------------------------------------------------------------*
		# MCOMMAND_CURRENTSTATETOOBJECT
		pop = self.Command_CSTO(op.GetDocument(), op)
		if pop is None:
			return None

		ndoc = c4d.documents.BaseDocument()
		if ndoc is None:
			return None
		nop = c4d.BaseObject(c4d.Onull)
		if nop is None:
			return None
		ndoc.InsertObject(nop)
		pop.InsertUnderLast(nop)

		# MCOMMAND_JOIN: Use temporary doc and Null object for grouping
		pop = self.Command_Join(ndoc, nop)
		if pop is None:
			return None
		# Get Global matrix transforms
		mg = op.GetUpMg()
		vadr = pop.GetAllPoints()
		if vadr is None:
			return None
		for i, obj in enumerate(vadr):
			vadr[i] = vadr[i] * mg

		# MCOMMAND_SUBDIVIDE
		if subd != 0:
			if self.Command_Subdivide(pop, subd) is False:
				return None

		pop.Remove()
		return pop
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Command_CSTO(self, cdoc, op):
	# - BaseDocument cdoc, BaseObject op
	#*---------------------------------------------------------------------------*
		# MCOMMAND_CURRENTSTATETOOBJECT
		res = utils.SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[op], doc=cdoc) 
		if res is False:
			return None
		elif not isinstance(res, list):
			return None
		return res[0]
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Command_Join(self, cdoc, op):
	# - BaseDocument cdoc, BaseObject op
	#*---------------------------------------------------------------------------*
		# MCOMMand_JOIN
		res = utils.SendModelingCommand(command=c4d.MCOMMAND_JOIN, list=[op], doc=cdoc) 
		if res is False:
			return None
		elif not isinstance(res, list):
			return None
		return res[0]
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Command_Subdivide(self, op, subd):
	# - BaseDocument BaseObject op, LONG subd
	#*---------------------------------------------------------------------------*
		# MCOMMand_SUBDIVIDE
		# Initialize Subdivide ModelingCommandData and BaseContainer
		sbc = c4d.BaseContainer()
		sbc[c4d.MDATA_SUBDIVIDE_HYPER] = False
		sbc[c4d.MDATA_SUBDIVIDE_ANGLE] = pi
		sbc[c4d.MDATA_SUBDIVIDE_SPLINESUB] = 0
		sbc[c4d.MDATA_SUBDIVIDE_SUB] = subd
		# Subdivide Mesh
		utils.SendModelingCommand(command=c4d.MCOMMAND_SUBDIVIDE, bc=sbc, list=[op], mode=c4d.MODELINGCOMMANDMODE_ALL, flags=c4d.MODELINGCOMMANDFLAGS_0)
		if res is None:
			return False
		return True
	#*---------------------------------------------------------------------------*

	# Calculate X-Z plane bounding rectangle for target's polygon
	#*---------------------------------------------------------------------------*
	def GetPolyBoundingRect_Tri(self, a, b, c):
	# - Vector a, b, c
	#*---------------------------------------------------------------------------*
		bounds = VMinMax()
		bounds.minX = bounds.maxX = a.x
		bounds.minZ = bounds.maxZ = a.z
		if b.x < bounds.minX:
			bounds.minX = b.x
		if b.x > bounds.maxX:
			bounds.maxX = b.x
		if b.z < bounds.minZ:
			bounds.minZ = b.z
		if b.z > bounds.maxZ:
			bounds.maxZ = b.z
		if c.x < bounds.minX:
			bounds.minX = c.x
		if c.x > bounds.maxX:
			bounds.maxX = c.x
		if c.z < bounds.minZ:
			bounds.minZ = c.z
		if c.z > bounds.maxZ:
			bounds.maxZ = c.z
		return bounds
	#*---------------------------------------------------------------------------*

	# Calculate X-Z plane bounding rectangle for target's polygon
	#*---------------------------------------------------------------------------*
	def GetPolyBoundingRect_Quad(self, a, b, c, d):
	# - Vector a, b, c, d
	#*---------------------------------------------------------------------------*
		bounds = VMinMax()
		bounds.minX = bounds.maxX = a.x
		bounds.minZ = bounds.maxZ = a.z
		if b.x < bounds.minX:
			bounds.minX = b.x
		if b.x > bounds.maxX:
			bounds.maxX = b.x
		if b.z < bounds.minZ:
			bounds.minZ = b.z
		if b.z > bounds.maxZ:
			bounds.maxZ = b.z
		if c.x < bounds.minX:
			bounds.minX = c.x
		if c.x > bounds.maxX:
			bounds.maxX = c.x
		if c.z < bounds.minZ:
			bounds.minZ = c.z
		if c.z > bounds.maxZ:
			bounds.maxZ = c.z
		if d.x < bounds.minX:
			bounds.minX = d.x
		if d.x > bounds.maxX:
			bounds.maxX = d.x
		if d.z < bounds.minZ:
			bounds.minZ = d.z
		if d.z > bounds.maxZ:
			bounds.maxZ = d.z
		return bounds
	#*---------------------------------------------------------------------------*

	# Drop object to floor
	#*---------------------------------------------------------------------------*
	def DropToFloor(self, ddoc, src, sop, offset):
	# - BaseDocument doc, BaseObject src, PolygonObject sop, LONG offset
	#*---------------------------------------------------------------------------*
		# Get Drop Value
		cnt = sop.GetPointCount()
		if cnt == 0:
			print "Error: DropIt.Process.DropToFloor.cnt"
			return True
		arr = sop.GetAllPoints()
		if arr is None:
			print "Error: DropIt.Process.DropToFloor.arr"
			return False
		# Find minimal Y value
		minY = arr[0].y
		for i, obj in enumerate(arr):
			if arr[i].y < minY:
				minY = arr[i].y

		# Drop Object to Floor
		minY -=	offset
		#print "Process.DropToFloor.minY-offset = ", minY
		# if the net movement is 0, do nothing
		if (abs(minY) < 0.000001):
			print "DropIt.Process.DropToFloor: No movement here"
			return False
		ddoc.AddUndo(c4d.UNDOTYPE_CHANGE, src)
		v = src.GetAbsPos()
		v.y -= minY
		src.SetAbsPos(v)
		# Notify object change and redraw
		src.Message(c4d.MSG_UPDATE)
		c4d.EventAdd()
		return True
	#*---------------------------------------------------------------------------*

	# Drop object to Workplane
	#*---------------------------------------------------------------------------*
	def DropToWorkplane(self, ddoc, src, sop, wpm, iwpm, offset):
	# - BaseDocument doc, BaseObject src, PolygonObject sop, Matrix wpm, Matrix iwpm, LONG offset
	#*---------------------------------------------------------------------------*
		# Get object points
		cnt = sop.GetPointCount()
		if cnt == 0:
			print "Error: DropIt.Process.DropToWorkplane.cnt"
			return True
		arr = sop.GetAllPoints()
		if arr is None:
			print "Error: DropIt.Process.DropToWorkplane.arr"
			return False
	
		# Transform sop points into workplane system
		# - Find minimal Y value in that system
		ov =	arr[0] * iwpm
		minY = ov.y
		for i, obj in enumerate(arr):
			ov =	arr[i] * iwpm
			if ov.y < minY:
				minY = ov.y

		# Drop Object to Floor
		minY -=	offset
		#print "DropIt.Process.DropToWorkplane.minY-offset = ", minY
		# if the net movement is 0, do nothing
		if (abs(minY) < 0.000001):
			print "DropIt.Process.DropToWorkplane: No movement here"
			return False
		ddoc.AddUndo(c4d.UNDOTYPE_CHANGE, src)
		# Transform drop object from global into workplane system, move, and back to global (right to left!)
		mg =	src.GetMg()
		src.SetMg(wpm * c4d.utils.MatrixMove(c4d.Vector(0.0,-minY,0.0)) * iwpm * mg)
		# Notify object change and redraw
		src.Message(c4d.MSG_UPDATE)
		c4d.EventAdd()
		return True
	#*---------------------------------------------------------------------------*

	# Drop object to target object
	#*---------------------------------------------------------------------------*
	def DropToTarget(self, doc, src, sop, dop, offset):
	# - BaseDocument doc, BaseObject src, PointObject sop, PolygonObject dop, LONG offset
	#*---------------------------------------------------------------------------*
		# Check Major Bounding Rectangles for overlap
		self.minY = BIGPOSREAL
		self.intersect = False
		if self.CheckMajorBoundingRects(sop, dop):
			# Find Closest distance between objects
			self.FindNearestIntersect(sop, dop)
		if self.intersect is False:
			MessageDialog("No intersection found")
			return True
		# Drop it! onto Target
		self.minY -= offset
		#print "Process.DropToTarget.minY = ", self.minY
		if abs(self.minY) < 0.000001:
			return True
		if abs(self.minY) > 9.0e10:
			return True
		doc.AddUndo(c4d.UNDOTYPE_CHANGE, src)
		v = src.GetAbsPos()
		v.y -= self.minY
		src.SetAbsPos(v)
		# Notify object change and redraw
		src.Message(c4d.MSG_UPDATE)
		c4d.EventAdd()
		return True
	#*---------------------------------------------------------------------------*

	# Compare X-Z plane bounding rectangles for overlap
	#*---------------------------------------------------------------------------*
	def CheckMajorBoundingRects(self, sop, dop):
	# - PointObject sop, PointObject dop
	#*---------------------------------------------------------------------------*
		# Source
		smm = self.GetBoundingRect(sop)
		if smm is None:
			return False
		# Target
		dmm = self.GetBoundingRect(dop)
		if dmm is None:
			return False

		# Check if bounding rectangles overlap
		# - If both min and max are beyond a side, no overlap
		if ((smm.minX < dmm.minX) and (smm.maxX < dmm.minX)):
			return False
		if ((smm.minX > dmm.maxX) and (smm.maxX > dmm.maxX)):
			return False
		if ((smm.minZ < dmm.minZ) and (smm.maxZ < dmm.minZ)):
			return False
		if ((smm.minZ > dmm.maxZ) and (smm.maxZ > dmm.maxZ)):
			return False
		return True
	#*---------------------------------------------------------------------------*

	# Calculate X-Z plane bounding rectangle for object
	#*---------------------------------------------------------------------------*
	def GetBoundingRect(self, op):
	# - PointObject op
	#*---------------------------------------------------------------------------*
		cnt = op.GetPointCount()
		if cnt == 0:
			print "Error: Process.GetBoundingRect.cnt"
			return None
		vadr = op.GetAllPoints()
		if vadr is None:
			print "Error: Process.GetBoundingRect.vadr"
			return None
		# Find minimal intersection
		vmm = VMinMax()
		for i,obj in enumerate(vadr):
				if (vadr[i].x < vmm.minX):
					vmm.minX = vadr[i].x
				if (vadr[i].x > vmm.maxX):
					vmm.maxX = vadr[i].x
				if (vadr[i].z < vmm.minZ):
					vmm.minZ = vadr[i].z
				if (vadr[i].z > vmm.maxZ):
					vmm.maxZ = vadr[i].z
		return vmm
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def FindNearestIntersect(self, sop, dop):
	# - PolygonObject sop
	#*---------------------------------------------------------------------------*
		# Check: Each object poly against every target poly (slow)
		padr =	sop.GetAllPolygons()
		if padr is None:
			print "Error: DropIt.Process.FindNearestIntersect.padr"
			return False
		vadr =	sop.GetAllPoints()
		if vadr is None:
			print "Error: DropIt.Process.FindNearestIntersect.vadr"
			return False
		dpoly =	dop.GetAllPolygons()
		if dpoly is None:
			print "Error: DropIt.Process.FindNearestIntersect.dpoly"
			return False
		dpoints =	dop.GetAllPoints()
		if dpoints is None:
			print "Error: DropIt.Process.FindNearestIntersect.dpoints"
			return False
		
		stri0 =	Triangle()
		stri1 =	Triangle()
		dtri =	Triangle()
		dist2 =	BIGPOSREAL
		# Loop thru Source polygons
		for i,obj in enumerate(padr):
			if (padr[i].c != padr[i].d):
				stri0.CopyFrom(vadr[padr[i].a],vadr[padr[i].b],vadr[padr[i].c],BIGNEGREAL)
				stri1.CopyFrom(stri0.c,vadr[padr[i].d],stri0.a,BIGNEGREAL)
				svmm = self.GetPolyBoundingRect_Quad(stri0.a, stri0.b, stri0.c, stri1.b)
				# Find minimal polygon intersection
				# Loop thru Target polygons
				for d,obj in enumerate(dpoly):
					# 1. Check if polygon bounding rectangles overlap
					# - If both min and max are beyond a side, no overlap
					if (svmm.minX < self.dbounds[d].minX) and (svmm.maxX < self.dbounds[d].minX):
						continue
					if (svmm.minX > self.dbounds[d].maxX) and (svmm.maxX > self.dbounds[d].maxX):
						continue
					if (svmm.minZ < self.dbounds[d].minZ) and (svmm.maxZ < self.dbounds[d].minZ):
						continue
					if (svmm.minZ > self.dbounds[d].maxZ) and (svmm.maxZ > self.dbounds[d].maxZ):
						continue
					# 2. Check for 2D triangle-triangle overlap
					# - Get minimum distance if there is overlap
					dtri.CopyFrom(dpoints[dpoly[d].a],dpoints[dpoly[d].b],dpoints[dpoly[d].c],BIGPOSREAL)
					if self.TrianglesOverlap(stri0, dtri):
						self.intersect =	True
						dist =		self.TriTriDistance(stri0, dtri)
						if (dist < dist2):
							dist2 = dist
					if self.TrianglesOverlap(stri1, dtri):
						self.intersect =	True
						dist =		self.TriTriDistance(stri1, dtri)
						if (dist < dist2):
							dist2 = dist
					if (dpoly[d].c == dpoly[d].d):
						continue
					dtri.CopyFrom(dtri.c,dpoints[dpoly[d].d],dtri.a,BIGPOSREAL)
					if self.TrianglesOverlap(stri0, dtri):
						self.intersect =	True
						dist =		self.TriTriDistance(stri0, dtri)
						if (dist < dist2):
							dist2 = dist
					if self.TrianglesOverlap(stri1, dtri):
						self.intersect =	True
						dist =		self.TriTriDistance(stri1, dtri)
						if (dist < dist2):
							dist2 = dist
			else:
				stri0.CopyFrom(vadr[padr[i].a],vadr[padr[i].b],vadr[padr[i].c],BIGNEGREAL)
				svmm = self.GetPolyBoundingRect_Tri(stri0.a, stri0.b, stri0.c)
				# Find minimal polygon intersection
				# Loop thru Target polygons
				for d,obj in enumerate(dpoly):
					# 1. Check if polygon bounding rectangles overlap
					# - If both min and max are beyond a side, no overlap
					if (svmm.minX < self.dbounds[d].minX) and (svmm.maxX < self.dbounds[d].minX):
						continue
					if (svmm.minX > self.dbounds[d].maxX) and (svmm.maxX > self.dbounds[d].maxX):
						continue
					if (svmm.minZ < self.dbounds[d].minZ) and (svmm.maxZ < self.dbounds[d].minZ):
						continue
					if (svmm.minZ > self.dbounds[d].maxZ) and (svmm.maxZ > self.dbounds[d].maxZ):
						continue
					# 2. Check for 2D triangle-triangle overlap
					# - Get minimum distance if there is overlap
					dtri.CopyFrom(dpoints[dpoly[d].a],dpoints[dpoly[d].b],dpoints[dpoly[d].c],BIGPOSREAL)
					if self.TrianglesOverlap(stri0, dtri):
						self.intersect =	True
						dist =				self.TriTriDistance(stri0, dtri)
						if (dist < dist2):
							dist2 = dist
					if (dpoly[d].c == dpoly[d].d):
						continue
					dtri.CopyFrom(dtri.c,dpoints[dpoly[d].d],dtri.a,BIGPOSREAL)
					if self.TrianglesOverlap(stri0, dtri):
						self.intersect =	True
						dist =				self.TriTriDistance(stri0, dtri)
						if (dist < dist2):
							dist2 = dist
		self.minY =	dist2
		return True
	#*---------------------------------------------------------------------------*

	#  Two dimensional Triangle-Triangle Overlap Test    
	#*---------------------------------------------------------------------------*
	def TrianglesOverlap(self, s, d):
	# - Triangle s, d
	#*---------------------------------------------------------------------------*
		if (self.Orient_2D(s.a,s.b,s.c) < 0.0):
			if (self.Orient_2D(d.a,d.b,d.c) < 0.0):
				return self.ccw_tri_tri_intersection_2d(s.a,s.c,s.b,d.a,d.c,d.b)
			else:
				return self.ccw_tri_tri_intersection_2d(s.a,s.c,s.b,d.a,d.b,d.c)
		elif (self.Orient_2D(d.a,d.b,d.c) < 0.0):
			return self.ccw_tri_tri_intersection_2d(s.a,s.b,s.c,d.a,d.c,d.b)
		return self.ccw_tri_tri_intersection_2d(s.a,s.b,s.c,d.a,d.b,d.c)
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def ccw_tri_tri_intersection_2d(self, sa, sb, sc, da, db, dc):
	# - Vector sa, sb, sc, da, db, dc
	#*---------------------------------------------------------------------------*
		if (self.Orient_2D(da,db,sa) >= 0.0):
			if (self.Orient_2D(db,dc,sa) >= 0.0):
				if (self.Orient_2D(dc,da,sa) >= 0.0):
					return True
				else:
					return self.intersection_test_edge(sa,sb,sc,da,db,dc)
			elif (self.Orient_2D(dc,da,sa) >= 0.0):
				return self.intersection_test_edge(sa,sb,sc,dc,da,db)
			else:
				return self.intersection_test_vertex(sa,sb,sc,da,db,dc)
		elif (self.Orient_2D(db,dc,sa) >= 0.0):
			if (self.Orient_2D(dc,da,sa) >= 0.0):
				return self.intersection_test_edge(sa,sb,sc,db,dc,da)
			else:
				return self.intersection_test_vertex(sa,sb,sc,db,dc,da)
		return self.intersection_test_vertex(sa,sb,sc,dc,da,db)
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def intersection_test_vertex(self, sa, sb, sc, da, db, dc):
	# - Vector sa, sb, sc, da, db, dc
	#*---------------------------------------------------------------------------*
		if (self.Orient_2D(dc,da,sb) >= 0.0):
			if (self.Orient_2D(dc,db,sb) <= 0.0):
				if (self.Orient_2D(sa,da,sb) > 0.0):
					if (self.Orient_2D(sa,db,sb) <= 0.0):
						return True
				elif (self.Orient_2D(sa,da,sc) >= 0.0):
					if (self.Orient_2D(sb,sc,da) >= 0.0):
						return True
			elif (self.Orient_2D(sa,db,sb) <= 0.0):
				if (self.Orient_2D(dc,db,sc) <= 0.0):
					if (self.Orient_2D(sb,sc,db) >= 0.0):
						return True
		elif (self.Orient_2D(dc,da,sc) >= 0.0):
			if (self.Orient_2D(sb,sc,dc) >= 0.0):
				if (self.Orient_2D(sa,da,sc) >= 0.0):
					return True
			elif (self.Orient_2D(sb,sc,db) >= 0.0):
				if (self.Orient_2D(dc,sc,db) >= 0.0):
					return True
		return False
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def intersection_test_edge(self, sa, sb, sc, da, db, dc):
	# - Vector sa, sb, sc, da, db, dc
	#*---------------------------------------------------------------------------*
		if (self.Orient_2D(dc,da,sb) >= 0.0):
			if (self.Orient_2D(sa,da,sb) >= 0.0):
				if (self.Orient_2D(sa,sb,dc) >= 0.0):
					return True
			elif (self.Orient_2D(sb,sc,da) >= 0.0):
				if (self.Orient_2D(sc,sa,da) >= 0.0):
					return True
		elif (self.Orient_2D(dc,da,sc) >= 0.0):
			if (self.Orient_2D(sa,da,sc) >= 0.0):
				if (self.Orient_2D(sa,sc,dc) >= 0.0):
					return True
				elif (self.Orient_2D(sb,sc,dc) >= 0.0):
					return True
		return False
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Orient_2D(self, a, b, c):
	# - Vector a, b, c
	#*---------------------------------------------------------------------------*
		return ((a.x-c.x)*(b.z-c.z)-(a.z-c.z)*(b.x-c.x))
	#*---------------------------------------------------------------------------*
	
	#*---------------------------------------------------------------------------*
	def TriTriDistance(self, s, d):
	# - Triangle s, d
	#*---------------------------------------------------------------------------*
		# Store minimum distance of Ray-Triangle intersections
		dist2 = BIGPOSREAL
		# Source.Target
		if self.RayTriangleIntersect(s.a, s.ra, d.a, d.b, d.c):
			dist = abs(s.a.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		if self.RayTriangleIntersect(s.b, s.rb, d.a, d.b, d.c):
			dist = abs(s.b.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		if self.RayTriangleIntersect(s.c, s.rc, d.a, d.b, d.c):
			dist = abs(s.c.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		if self.RayTriangleIntersect(s.ctr, s.rctr, d.a, d.b, d.c):
			dist = abs(s.ctr.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		# Target.Source
		if self.RayTriangleIntersect(d.a, d.ra, s.a, s.b, s.c):
			dist = abs(d.a.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		if self.RayTriangleIntersect(d.b, d.rb, s.a, s.b, s.c):
			dist = abs(d.b.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		if self.RayTriangleIntersect(d.c, d.rc, s.a, s.b, s.c):
			dist = abs(d.c.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		if self.RayTriangleIntersect(d.ctr, d.rctr, s.a, s.b, s.c):
			dist = abs(d.ctr.y-self.yhitpos)
			if (dist < dist2):
				dist2 = dist
		return dist2
	#*---------------------------------------------------------------------------*

	# Process.RayTriangleIntersect
	#*---------------------------------------------------------------------------*
	def RayTriangleIntersect(self, p0, p1, lv0, lv1, lv2):
	# - Vector p0, p1, lv0, lv1, lv2
	#*---------------------------------------------------------------------------*
		#GePrint("RayTriangleIntersect("+VectorToString(p0)+", "+VectorToString(p1)+", "+VectorToString(lv0)+", "+VectorToString(lv1)+", "+VectorToString(lv2)+")");
		# get triangle edge vectors and plane normal
		u = lv1 - lv0
		v = lv2 - lv0
		# cross product
		n = u % v
		# triangle is degenerate
		if c4d.utils.VectorEqual(n, c4d.Vector(0.0)):
			#print "DropIt.Process.RayTriangleIntersect.Degenerate Triangle"
			return False
			# do not deal with this case
	
		# ray vectors
		# params to calc ray-plane intersect
		# - ray direction vector
		dir = p1 - p0
		b = n * dir
		# ray is parallel to triangle plane
		if (abs(b) < 0.00000001):
			# ray lies in triangle plane
			#if (a == 0)	return FALSE;
			# ray disjoint from plane
			#print "DropIt.Process.RayTriangleIntersect.Ray Parallel Triangle Plane"
			return False

		a =	-(n * (p0 - lv0))
	
		# get intersect point of ray with triangle plane
		r = a / b
		# ray goes away from triangle
		if (r < 0.0):
			#print "DropIt.Process.RayTriangleIntersect.No Intersect"
			# => no intersect
			return False
		# for a segment, also test if (r > 1.0) => no intersect
	
		# intersect point of ray and plane
		hitpos = p0 + r * dir
	
		# is I inside T?
		nn = hitpos - lv0
		wu = nn * u
		wv = nn * v
		uv = u * v
		uu = u * u
		vv = v * v
		D = 1.0 / ((uv * uv) - (uu * vv))
	
		# get and test parametric coords
		ta = ((uv * wv) - (vv * wu)) * D
		# I is outside T
		if ((ta < 0.0) or (ta > 1.0)):
			#print "DropIt.Process.RayTriangleIntersect.I is outside T.a"
			return False
		tb = ((uv * wu) - (uu * wv)) * D
		# I is outside T
		if ((tb < 0.0) or ((ta + tb) > 1.0)):
			#print "DropIt.Process.RayTriangleIntersect.I is outside T.b"
			return False
	
		# I is in T
		self.yhitpos =	hitpos.y
		return True
	#*---------------------------------------------------------------------------*
#*****************************************************************************

#*****************************************************************************
class DIDialog(gui.GeDialog):
	# Data
	# - BaseDocument
	sdoc = None
	ddoc = None
	# - Base2DList
	src = None
	# - Base2DList
	dst = None
	# - Real
	offset = 0.0
	# - Long
	target = TARGET_FLOOR
	ssubd = 0
	dsubd = 0
	width = 320
	height = 0

	#*---------------------------------------------------------------------------*
	def CreateLayout(self):
	#*---------------------------------------------------------------------------*
		#set the title
		self.SetTitle("Drop It! v1.6")
		# .
		self.GroupBegin(id=0,flags=c4d.BFH_SCALEFIT|c4d.BFV_SCALEFIT,cols=1,rows=8,title="",groupflags=0)
		self.GroupBorderSpace(4,4,4,4)
		# -.
		self.GroupBegin(id=0,flags=c4d.BFH_SCALEFIT,cols=1,rows=2,title="Drop Objects",groupflags=0)
		self.GroupBorder(c4d.BORDER_WITH_TITLE|c4d.BORDER_THIN_IN)
		self.GroupBorderSpace(4,4,4,4)
		self.GroupBegin(id=0,flags=c4d.BFH_CENTER,cols=2,rows=1,title="",groupflags=0)
		self.AddButton(id=DID_SOURCE,flags=c4d.BFH_CENTER|c4d.BFV_CENTER,initw=0,inith=0,name="Drop Objects")
		self.AddCheckbox(id=DID_SOURCE_SET,flags=c4d.BFH_CENTER|c4d.BFV_CENTER,initw=0,inith=0,name="")
		self.GroupEnd();
		self.AddStaticText(id=0,flags=c4d.BFH_LEFT|c4d.BFV_CENTER,initw=0,inith=0,name="Select Drop Objects, Click 'Drop Objects' button ",borderstyle=0)
		self.GroupEnd()
		# -.
		self.GroupBegin(id=0,flags=c4d.BFH_SCALEFIT,cols=1,rows=3,title="Target",groupflags=0)
		self.GroupBorder(c4d.BORDER_WITH_TITLE|c4d.BORDER_THIN_IN)
		self.GroupBorderSpace(4,4,4,4)
		self.AddComboBox(id=DID_WHICHTARGET,flags=c4d.BFH_LEFT|c4d.BFV_CENTER)
		self.AddChild(id=DID_WHICHTARGET,subid=TARGET_FLOOR,child="Floor")
		if c4d.GetC4DVersion() > 14033:
			self.AddChild(id=DID_WHICHTARGET,subid=TARGET_WORKPLANE,child="Workplane")
		self.AddChild(id=DID_WHICHTARGET,subid=TARGET_OBJECT,child="Object(s)")
		self.GroupBegin(id=0,flags=c4d.BFH_CENTER,cols=2,rows=1,title="",groupflags=0)
		self.AddButton(id=DID_TARGET,flags=c4d.BFH_CENTER|c4d.BFV_CENTER,initw=0,inith=0,name="Target Objects")
		self.AddCheckbox(id=DID_TARGET_SET,flags=c4d.BFH_CENTER|c4d.BFV_CENTER,initw=0,inith=0,name="")
		self.GroupEnd();
		self.AddStaticText(id=0,flags=c4d.BFH_LEFT|c4d.BFV_CENTER,initw=0,inith=0,name="Select Target Objects, Click 'Target Objects' button ",borderstyle=0)
		self.GroupEnd()
		# -.
		self.GroupBegin(id=0,flags=c4d.BFH_SCALEFIT,cols=2,rows=1,title="",groupflags=0)
		self.AddStaticText(id=0,flags=c4d.BFH_LEFT|c4d.BFV_CENTER,initw=0,inith=0,name="Object Subdivision: ",borderstyle=0)
		self.AddEditNumberArrows(id=DID_SSUBD,flags=c4d.BFH_SCALEFIT,initw=0,inith=0)
		self.GroupEnd()
		# -.
		self.GroupBegin(id=0,flags=c4d.BFH_SCALEFIT,cols=2,rows=1,title="",groupflags=0)
		self.AddStaticText(id=0,flags=c4d.BFH_LEFT|c4d.BFV_CENTER,initw=0,inith=0,name="Target Subdivision: ",borderstyle=0)
		self.AddEditNumberArrows(id=DID_DSUBD,flags=c4d.BFH_SCALEFIT,initw=0,inith=0)
		self.GroupEnd()
		# -.
		self.GroupBegin(id=0,flags=c4d.BFH_SCALEFIT,cols=2,rows=1,title="",groupflags=0)
		self.AddStaticText(id=0,flags=c4d.BFH_LEFT|c4d.BFV_CENTER,initw=0,inith=0,name="Offset: ",borderstyle=0)
		self.AddEditNumberArrows(id=DID_OFFSET,flags=c4d.BFH_SCALEFIT,initw=0,inith=0)
		self.GroupEnd()
		# .
		self.AddSeparatorH(inith=0)
		self.AddButton(id=DID_DROPIT,flags=c4d.BFH_CENTER|c4d.BFV_CENTER,initw=0,inith=0,name="Drop It!")
		self.GroupEnd()
		return True
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def InitValues(self):
	#*---------------------------------------------------------------------------*
		self.SetReal(id=DID_OFFSET,value=self.offset,step=0.1)
		self.SetLong(id=DID_SSUBD,value=self.ssubd,min=0,max=8)
		self.SetLong(id=DID_DSUBD,value=self.dsubd,min=0,max=8)
		self.SetLong(id=DID_WHICHTARGET,value=self.target)
		self.SetBool(id=DID_SOURCE_SET,value=False)
		self.SetBool(id=DID_TARGET_SET,value=False)
		self.Enable(DID_SOURCE_SET,False)
		self.Enable(DID_TARGET_SET,False)
		return True
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Message(self, msg, result):
	# - BaseContainer msg, result
	#*---------------------------------------------------------------------------*
		if msg.GetId() == c4d.BFM_ADJUSTSIZE:
			self.width = msg.GetLong(c4d.BFM_ADJUSTSIZE_WIDTH)
			self.height = msg.GetLong(c4d.BFM_ADJUSTSIZE_HEIGHT)
		return gui.GeDialog.Message(self, msg, result)
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Command(self, id, msg):
	# - LONG id, BaseContainer msg
	#*---------------------------------------------------------------------------*
		if id == DID_SOURCE:
			self.sdoc = c4d.documents.GetActiveDocument()
			self.src = self.sdoc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN)
			if len(self.src):
				self.SetBool(id=DID_SOURCE_SET,value=True)
			else:
				self.SetBool(id=DID_SOURCE_SET,value=False)
			#for i, obj in enumerate(self.src):     #Loop through the selected objects
				#print self.src[i].GetName()
		elif id == DID_TARGET:
			self.target = self.GetLong(DID_WHICHTARGET)
			# Only get targets if not using Floor or Workplane
			if self.target != TARGET_OBJECT:
				return True
			self.ddoc = c4d.documents.GetActiveDocument()
			# No, you can't drop objects from one document onto another document's objects
			if self.ddoc != self.sdoc:
				c4d.gui.MessageDialog("This is not the same document as your drop object(s)!")
				return True
			self.dst = self.ddoc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN)
			if len(self.dst):
				self.SetBool(id=DID_TARGET_SET,value=True)
			else:
				self.SetBool(id=DID_TARGET_SET,value=False)
			#for i, obj in enumerate(self.dst):     #Loop through the selected objects
				#print self.dst[i].GetName()
		elif id == DID_DROPIT:
			if not (self.src and len(self.src)):
				c4d.gui.MessageDialog("No drop objects have been added.")
				return True
			# Get Offset and Subdivisions - Numeric value
			self.offset = self.GetReal(DID_OFFSET)
			self.ssubd = self.GetLong(DID_SSUBD)
			self.dsubd = self.GetLong(DID_DSUBD)
			self.target = self.GetLong(DID_WHICHTARGET)
			# Drop It!
			c4d.StopAllThreads()
			c4d.gui.SetMousePointer(c4d.MOUSE_BUSY)
			process = Process()
			if self.target == TARGET_FLOOR:
				process.Start_Floor(self.sdoc, self.src, self.offset, self.ssubd)
			elif self.target == TARGET_WORKPLANE:
				process.Start_Workplane(self.sdoc, self.src, self.offset, self.ssubd)
			elif not len(self.dst) is 0:
				process.Start_Object(self.sdoc, self.src, self.dst, self.offset, self.ssubd, self.dsubd)
			else:
				c4d.gui.MessageDialog("No target objects have been added.")
			# Clear source and destination object lists, clear checkboxes
			self.src = []
			self.dst = []
			self.SetBool(id=DID_SOURCE_SET,value=False)
			self.SetBool(id=DID_TARGET_SET,value=False)
			c4d.gui.SetMousePointer(c4d.MOUSE_NORMAL)
		return True
	#*---------------------------------------------------------------------------*
#*****************************************************************************

#*****************************************************************************
class DropIt(plugins.CommandData):

	# Data
	dialog = None

	#*---------------------------------------------------------------------------*
	def GetState(self, doc):
	# - BaseDocument doc
	#*---------------------------------------------------------------------------*
		return c4d.CMD_ENABLED
	#*---------------------------------------------------------------------------*

	#*---------------------------------------------------------------------------*
	def Execute(self, doc):
	# - BaseDocument doc
	#*---------------------------------------------------------------------------*
		# Create DropIt non-modal Dialog
		if self.dialog is None:
			self.dialog = DIDialog()
		if self.dialog.IsOpen():
			return True
		return self.dialog.Open(dlgtype=c4d.DLG_TYPE_ASYNC, pluginid=PLUGIN_ID, xpos=-1, ypos=-1, defaultw=self.dialog.width, defaulth=self.dialog.height,subid=ID_DIDIALOG)
	#*---------------------------------------------------------------------------*
#*****************************************************************************


if __name__ == "__main__":
	bmp = bitmaps.BaseBitmap()
	dir, f = os.path.split(__file__)
	fn = os.path.join(dir, "res", "pluginicon.tif")
	bmp.InitWith(fn)

	# Console Banner
	print " "
	print "=========================="
	print "-- DropIt! v1.6 "+u"\u00a9"+" 2015"
	print "=========================="
	print " "

	c4d.plugins.RegisterCommandPlugin(id=PLUGIN_ID,str="Drop It! (Python)",info=0,help="Place objects onto floor, workplane, or target object",dat=DropIt(),icon=bmp)
